library(mice)
library(mvtnorm)
library(Amelia)
library(missMDA)   
source("Estim_param.r")
source("BayesMIPCA.r")

concat<-function(X,collaps="",sepa="",ext=TRUE,roc=3){
# PURPOSE: Concatenate the different element of a vector or a matrix,
# with the possibility to do it in columns or in rows
# C. CARPENTIER (2009)
  if( is.null( dim(X) ) ) {
    if (ext==TRUE) { ex =collaps } else { ex =""}
    paste(ex,paste(X,collapse=collaps,sep=sepa),ex,sep=sepa)
  } else {
    if (roc ==3) {
      Y <- as.vector(X)
      concat(Y,collaps,sepa,ext)
    } else { apply(X, roc, concat, collaps,sepa, ext) }
  } 
}

createdata <- function(n, Sigma) {
# PURPOSE: generate a data set
  X<-rmvnorm(n,sigma=Sigma)
  colnames(X)<-c("y",paste("x",1:(ncol(Sigma)-1),sep=""))
  return(as.data.frame(X))
}

Missing<-function(don,p=0.1,S){
# PURPOSE: generate a proportion p of missing values
# completely at random while insuring not to much
# missing values on a same line
  convert.missing<-function(p,P,S,pas=1/1000){
    esp<-function(p,P,S){
      Max<-(P-S)
      s<-sum(dbinom((Max+1):P,P,p))
      esp<-sum(dbinom(1:Max,P,p)*(1/(1-s))*1:Max)
      return(esp)
    }
    for(i in seq(0,1,pas)){if((esp(i,P,S)/P)>p){return(i)}}
  }
  p.conv<-convert.missing(p=p,P=ncol(don),S=S)
  don.na<-don
  while(sum(is.na(don.na))==0){
    nb.na<-ncol(don)
    pattern<-matrix(FALSE,nrow(don),ncol(don))
    pos.na<-Inf
    for (i in 1:nrow(don)){
      while (sum(pos.na)>(ncol(don)-S)){
	pos.na<-rbinom(n=nb.na,size=1,p=p.conv)
      }
    pattern[i,]<-as.logical(pos.na)
    pos.na<-Inf
    }
    don.na[pattern]<-NA
  }
  return(don.na)
}

Transform.mice<-function(res.MI,data){
#PURPOSE : modifiy the Amelia's and BayesMIPCA 's output
# to analyse the imputed data sets using mice's functions
  if(class(res.MI)=="amelia"){
    m<-res.MI$m
    temp<-array(NA,dim=c(dim(res.MI[["imputations"]][["imp1"]]),m))
    for (i in 1:m){temp[,,i]<-as.matrix(res.MI[["imputations"]][[i]])} 
    resMI<-temp 
  }else if(class(res.MI)=="BayesMIPCA"){
    resMI<-res.MI
    m<-length(resMI)
    temp<-array(NA,dim=c(dim(data),m))
    for (i in 1:m){temp[,,i]<-as.matrix(resMI[[i]])}
    resMI<-temp
  }
  imp<-c();imp$imp<-as.list(colnames(data))
  missing<-apply(is.na(data),MARGIN=2,FUN="which")
  temp<-apply(resMI,MARGIN=3,FUN=function(don,missing){l<-c();for(i in 1:ncol(don)){l<-c(l,list(don[missing[[i]],i]))};names(l)<-colnames(don);return(l)},missing=missing)
  for (var in 1:ncol(data)){
    if (length(missing[[var]])==0){imp$imp[var]<-list(NULL)}else{
      temp2<-c()
      for (num in 1:m){
	temp2<-cbind(temp2,temp[[num]][[var]])
      }
      imp$imp[[var]]<-temp2
    }
  }
  names(imp$imp)<-colnames(data)
  imp$m<-m
  imp$call<-NA
  imp$nmis<-apply(data,MARGIN=2,FUN=function(x){sum(is.na(x))})
  imp$data<-data
  class(imp)<-"mids"
  return(imp)
}

ddl<-function(U,B,nucom,m){
#PURPOSE: calculate the degrees of freedom
# for the confidence interval according to
# the Barnard and Rubin's method
  lambda<-(B+B/m)/(U+(B+B/m))
  nuobs<-((nucom+1)/(nucom+3))*nucom*(1-lambda)
  nuold<-(m-1)/lambda^2
  nu<-nuold*nuobs/(nuold+nuobs)
  return(nu)
}
  
cor.mice<-function(imp,v1.cor,v2.cor){
# PURPOSE: calculate the pooled estimate
# and the confidence interval for the regression coefficient
# using the Fisher z transformation
  dataset<-imp$data[,c(v1.cor,v2.cor)]
  Tab<-array(NA,dim=c(nrow(imp$data),2,imp[["m"]]),dimnames=list(rownames(imp$data),colnames(dataset),as.character(1:imp$m)))
  bool<-apply(is.na(as.matrix(dataset)),2,any)
  for (i in 1:imp$m){
    Tab[,,i]<-as.matrix(dataset)
    if(bool[1]){Tab[which(is.na(dataset[,colnames(dataset)[1]])),1,i]<-imp[["imp"]][[colnames(dataset)[1]]][,i]}
    if(bool[2]){Tab[which(is.na(dataset[,colnames(dataset)[2]])),2,i]<-imp[["imp"]][[colnames(dataset)[2]]][,i]}
  }
  rho<-apply(Tab,MARGIN=c(3),FUN=function(x){cor(x)[1,2]})
  z<-.5*log((1+rho)/(1-rho))
  var_intra<-1/(nrow(imp$data)-3)
  var_inter<-var(z)
  se<-sqrt(var_intra+(1+1/imp$m)*var_inter)
  nucom<-nrow(imp$data)-1
  nu<-ddl(var_intra,var_inter,nucom,imp[["m"]])
  zbar<-mean(z)
  Binf<-zbar-qt(.975,nu)*se
  Bsup<-zbar+qt(.975,nu)*se
  res<-sapply(list(est=zbar,"lo 95"=Binf,"hi 95"=Bsup),tanh)
  return(res)
}

LD<-function(data,formula,v1.cor,v2.cor){
# PURPOSE: calculate the pooled estimate
# and the confidence interval for the three quantities of interest
# using listwise deletion
  imp<-na.omit(data)
  #beta
  res.lm<-lm(as.character(formula),data=imp)
  res.beta<-c(summary(res.lm)[["coefficients"]]["x1",c("Estimate")],confint(res.lm)["x1",])
  names(res.beta)<-c("est","lo 95","hi 95")
  # mean
  res.lm<-lm(y~1,data=imp)
  res.mean<-c(summary(res.lm)[["coefficients"]][,c("Estimate")],confint(res.lm))
  names(res.mean)<-c("est","lo 95","hi 95")
  # cor
  rho<-cor(imp[,v1.cor],imp[,v2.cor])
  z<-.5*log((1+rho)/(1-rho))
  Binf<-qnorm(.975)*sqrt(1/(nrow(imp)-3))
  Bsup<-Binf
  Binf<-tanh(z-Binf)
  Bsup<-tanh(z+Bsup)
  res.cor<-c(rho,Binf,Bsup);names(res.cor)<-c("est","lo 95","hi 95")
  return(list(beta=res.beta,mean=res.mean,cor=res.cor))
}

test.impute <- function(data, method, formula, m=20,S,L=100, Lstart=1000, verbose=F,kapa,empri,predictorMatrix,v1.cor,v2.cor,...){
# PURPOSE: calculate the pooled estimate
# and the confidence interval for the three quantities of interest
# using a multiple imputation method
# (van Buuren 2012)
    print(method)
    if (method=="BayesMIPCA"){
    imp<-Transform.mice(BayesMIPCA(data,M=m,S=S,L=L,Lstart=Lstart, verbose=verbose),data)
    }else if(method=="amelia"){imp<-Transform.mice(amelia(data,m=m,p2s=if(verbose){1}else{0},empri=empri),data)
    }else if(method=="LD"){return(LD(data=data,formula=formula,v1.cor=v1.cor,v2.cor=v2.cor))
    }else{imp <- mice(data, method=method, m=m, print=verbose, ridge=kapa,predictorMatrix=predictorMatrix,...)} 
    fit <- list(with(imp, lm(as.formula(as.character(formula)))),with(imp, lm(y~1)))   
    est <- list(pool(fit[[1]]),pool(fit[[2]]))
    tab <- list(summary(est[[1]]),summary(est[[2]]))
    tab[[3]]<-cor.mice(imp,v1.cor,v2.cor)
    return(list(beta=tab[[1]]["x1",c("est","lo 95","hi 95")],mean=tab[[2]][1,c("est","lo 95","hi 95")],cor=tab[[3]]))
}


simulate <- function(nsim=10,p,m=20,Sigma.create,n,formula, S, verbose=F, methods, kapa=0,predictorMatrix=(1 - diag(1, ncol(data))),empri=0, L=100, Lstart=1000,v1.cor,v2.cor,...){
# PURPOSE: put together all the results for several
# incomplete data sets
# (van Buuren 2012)
  res_beta <- array(NA,dim=c(4, nsim, 3))
  res_mean <- array(NA,dim=c(4, nsim, 3))
  res_cor <- array(NA,dim=c(4, nsim, 3))
  dimnames(res_beta) <- list(c("LD","amelia","BayesMI","BayesMIPCA"), as.character(1:nsim),c("est","lo 95","hi 95"))
  dimnames(res_mean)<-dimnames(res_cor)<-dimnames(res_beta)
  im <- c("LD","amelia","norm","BayesMIPCA")
  for(i in 1:nsim){
    print(i)
    set.seed(i)
    data <- createdata(n=n,Sigma=Sigma.create)
    data <- Missing(data,p=p,S=S)
    if(1%in%methods){
      temp<-test.impute(data, method=im[1],formula=formula,v1.cor=v1.cor,v2.cor=v2.cor)
      res_beta[1,i,] <- temp[[1]]
      res_mean[1,i,] <- temp[[2]]
      res_cor[1,i,] <- temp[[3]]
    }
    if(2%in%methods){
      temp<-test.impute(data, method=im[2],m=m,empri=empri,formula=formula,verbose=verbose,v1.cor=v1.cor,v2.cor=v2.cor)
      res_beta[2,i,] <- temp[[1]]
      res_mean[2,i,] <- temp[[2]]
      res_cor[2,i,] <- temp[[3]]
    }
    if (3%in%methods){
      temp<-test.impute(data, method=im[3],m=m,kapa=kapa,formula=formula,verbose=verbose,predictorMatrix=predictorMatrix,v1.cor=v1.cor,v2.cor=v2.cor)
      res_beta[3,i,] <- temp[[1]]
      res_mean[3,i,] <- temp[[2]]
      res_cor[3,i,] <- temp[[3]]
    }
    if(4%in%methods){
      temp<-test.impute(data, method=im[4],S=S,formula=formula,m=m,verbose=verbose,L=L,Lstart=Lstart,v1.cor=v1.cor,v2.cor=v2.cor)
      res_beta[4,i,] <- temp[[1]]
      res_mean[4,i,] <- temp[[2]]
      res_cor[4,i,] <- temp[[3]]
    }

  }
  return(list(beta=res_beta,mean=res_mean,cor=res_cor))
}

summarize.results <- function(res,true_beta,true_mean,true_cor,digits=4){
# PURPOSE: evaluate the quality of imputation using the several criteria 
# (van Buuren 2012)
  # bias of beta1, mean y, cor
  bias_beta <- rowMeans(res$beta[,,1] - true_beta)
  bias_mean <- rowMeans(res$mean[,,1] - true_mean)   
  bias_cor <- rowMeans(res$cor[,,1] - true_cor)
  
  # coverage
  isin_beta <- res$beta[,,2] < true_beta & true_beta < res$beta[,,3]
  cov_beta <- rowMeans(isin_beta)
  isin_mean <- res$mean[,,2] < true_mean & true_mean < res$mean[,,3]
  cov_mean <- rowMeans(isin_mean)
  isin_cor <- res$cor[,,2] < true_cor & true_cor < res$cor[,,3]
  cov_cor <- rowMeans(isin_cor,na.rm=T)

  # average width of 95 c.i.
  intwidth_beta <- res$beta[,,3] - res$beta[,,2]
  aiw_beta<-apply(intwidth_beta,1,median)
  intwidth_mean <- res$mean[,,3] - res$mean[,,2]
  aiw_mean <- apply(intwidth_mean,1,median)
  intwidth_cor <- res$cor[,,3] - res$cor[,,2]
  aiw_cor<-apply(intwidth_cor,1,median,na.rm=T)
  
  results<-lapply(list(beta=rbind(biais=na.omit(bias_beta),aiw=na.omit(aiw_beta),cov=na.omit(cov_beta),rmse=sqrt(na.omit(rowMeans((res$beta[,,1] - true_beta)^2)))),
  mean=rbind(biais=na.omit(bias_mean),aiw=na.omit(aiw_mean),cov=na.omit(cov_mean),rmse=sqrt(na.omit(rowMeans((res$mean[,,1] - true_mean)^2)))),
  cor=rbind(biais=na.omit(bias_cor),aiw=na.omit(aiw_cor),cov=na.omit(cov_cor),rmse=sqrt(na.omit(rowMeans((res$cor[,,1] - true_cor)^2))))
  ),round,digits=digits)
  return(results)
}